@@ -13,7 +13,6 @@
 
 import operator
 import re
-from six import moves
 
 from .. import JSONPath, DatumInContext, Index
 
@@ -26,7 +25,7 @@ OPERATOR_MAP = {
     '<': operator.lt,
     '>=': operator.ge,
     '>': operator.gt,
-    '=~': lambda a, b: True if re.search(b, a) else False,
+    '=~': lambda a, b: True if isinstance(a, str) and re.search(b, a) else False,
 }
 
 
@@ -49,11 +48,21 @@ class Filter(JSONPath):
             return []
 
         return [DatumInContext(datum.value[i], path=Index(i), context=datum)
-                for i in moves.range(0, len(datum.value))
+                for i in range(0, len(datum.value))
                 if (len(self.expressions) ==
                     len(list(filter(lambda x: x.find(datum.value[i]),
                                     self.expressions))))]
 
+    def filter(self, fn, data):
+        # NOTE: We reverse the order just to make sure the indexes are preserved upon
+        #  removal.
+        for datum in reversed(self.find(data)):
+            index_obj = datum.path
+            if isinstance(data, dict):
+                index_obj.index = list(data)[index_obj.index]
+            index_obj.filter(fn, data)
+        return data
+
     def update(self, data, val):
         if type(data) is list:
             for index, item in enumerate(data):
@@ -95,16 +104,11 @@ class Expression(JSONPath):
         found = []
         for data in datum:
             value = data.value
-            if isinstance(self.value, int):
+            if type(self.value) is int:
                 try:
                     value = int(value)
                 except ValueError:
                     continue
-            elif isinstance(self.value, bool):
-                try:
-                    value = bool(value)
-                except ValueError:
-                    continue
 
             if OPERATOR_MAP[self.op](value, self.value):
                 found.append(data)
