@@ -1,7 +1,7 @@
 import sys
 import logging
 
-import ply.lex
+import jsonpath_ng._ply.lex
 
 from jsonpath_ng.exceptions import JsonPathLexerError
 
@@ -23,7 +23,7 @@ class JsonPathLexer:
         Maps a string to an iterator over tokens. In other words: [char] -> [token]
         '''
 
-        new_lexer = ply.lex.lex(module=self, debug=self.debug, errorlog=logger)
+        new_lexer = jsonpath_ng._ply.lex.lex(module=self, debug=self.debug, errorlog=logger)
         new_lexer.latest_newline = 0
         new_lexer.string_value = None
         new_lexer.input(string)
@@ -64,7 +64,9 @@ class JsonPathLexer:
     t_ignore = ' \t'
 
     def t_ID(self, t):
-        r'[a-zA-Z_@][a-zA-Z0-9_@\-]*'
+        # CJK: [\u4E00-\u9FA5]
+        # EMOJI: [\U0001F600-\U0001F64F]
+        r'([a-zA-Z_@]|[\u4E00-\u9FA5]|[\U0001F600-\U0001F64F])([a-zA-Z0-9_@\-]|[\u4E00-\u9FA5]|[\U0001F600-\U0001F64F])*'
         t.type = self.reserved_words.get(t.value, 'ID')
         return t
 
